"
 COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

ArrayedCollection subclass:#ExternalBytes
       instanceVariableNames:'address base size'
       classVariableNames:''
       poolDictionaries:''
       category:'System-Support'
!

ExternalBytes comment:'
COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

$Header: /claus/smalltalk/st/libbasic/RCS/ExtBytes.st,v 1.9 95/02/15 10:21:53 claus Exp $
'!

!ExternalBytes class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /claus/smalltalk/st/libbasic/RCS/ExtBytes.st,v 1.9 95/02/15 10:21:53 claus Exp $
"
!

documentation
"
    This class provides access to any memory in the system. Its main purpose
    is to provide a base for classes referencing external data. Use with
    great care - access is not checked for out-of-bounds or valid addresses.
    Also notice, that this class is not available on other smalltalk systems,
    making code using it very unportable.
    Since the data is allocated outside the garbage collected smalltalk space,
    its address stays fix. Thus, it can be passed to external C-functions without
    any danger. However, you have to take care for freeing the memory yourself.

    Example:
        |buffer|

        buffer := ExternalBytes new:100.
        ...
        buffer free
"
! !

!ExternalBytes class methodsFor:'instance creation'!

address:aNumber
    "return a new ExternalBytes object to access bytes starting at aNumber.
     The memory at aNumber has been allocated elsewhere. The size is not known."

    ^ self basicNew setAddress:aNumber
!

new:numberOfBytes
    "allocate some memory usable for data;
     return a corresponding ExternalBytes object or nil (if malloc fails)"

    ^ self newForData:numberOfBytes
!

newForText:numberOfBytes
    "allocate some memory usable to put machine executable instructions into;
     return a corresponding ExternalBytes object or nil. The data will be
     aligned so that loading an object files text into it is possible.
     These bytes are NOT subject to garbage collection; therefore someone
     must take care of freeing the memory when its no longer needed.
     This method may not be available on all architectures (i.e. those that
     do not allow allocating text-memory)."

    ^ self newForText:numberOfBytes align:true
!

newForText:numberOfBytes align:align
    "allocate some memory usable to put machine executable instructions into;
     return a corresponding ExternalBytes object or nil (if malloc fails).
     If align is set to true, the data will be aligned so that loading of
     text from an object file will be possible (not all machines need this).
     These bytes are NOT subject to garbage collection; therefore someone
     must take care of freeing the memory when its no longer needed.
     This method may not be available on all architectures (i.e. those that
     do not allow allocating text-memory)."

    |addr baseAddr|
%{
#ifdef mips
# include <scnhdr.h>
#endif

#ifndef SCNROUND
# define SCNROUND 0
#endif

#if defined(CAN_USE_DATA_FOR_TEXT)
    char *space;
    unsigned long _addr = 0;
    unsigned long alignment = SCNROUND;

    if (align != true)
        alignment = 0;

    /*
     * this one should allocate some space, into which we can put
     * some code to be executed. You are in bad luck, if running on
     * a SepId machine or if data space cannot be executed.
     * all you can do then, is to look for any OS-hooks.
     */
    if (__isSmallInteger(numberOfBytes)) {
        space = (char *) malloc(_intVal(numberOfBytes) + alignment);
        _addr = (unsigned long) space;
    }

    if (_addr <= 0) {
        RETURN ( nil );
    }
    if (alignment) {
        baseAddr = _MKSMALLINT(_addr);
        _addr &= ~(alignment - 1);
        _addr += alignment;
# ifdef SUPERDEBUG
        printf("rounding address to: %x (%d-aligned)\n", _addr, alignment);
# endif
    }
    addr = _MKSMALLINT(_addr);
#endif
%}.
    ^ self basicNew setAddress:addr base:baseAddr size:numberOfBytes
!

newForData:numberOfBytes
    "allocate some memory usable for data;
     return a corresponding ExternalBytes object or nil (if malloc fails)"

    |addr|
%{
    char *space;
    unsigned long _addr = 0;

    /*
     * this one is not at all critical - simply return some space
     * to be used for data space of loaded-in binaries.
     */
    if (__isSmallInteger(numberOfBytes)) {
        space = (char *) malloc(_intVal(numberOfBytes));
        _addr = (unsigned long) space;
    }
    if (_addr <= 0) {
        RETURN (nil);
    }

    addr = _MKSMALLINT(_addr);
%}.
    ^ self basicNew setAddress:addr size:numberOfBytes
! !

!ExternalBytes methodsFor:'freeing'!

free
    "free a previously allocated piece of memory - be very careful, there
     are no checks done here. All dangers you usually have with malloc/free
     are present here ..."

    "at least, we check for double freeing the same chunk"
    address isNil ifTrue:[
        self error:'freeing memory twice'.
        ^ self
    ].
%{
    OBJ b;

    b = _INST(base);
    if (b != nil) {
        free((char *)(_intVal(b)));
    } else {
        free((char *)(_intVal(_INST(address))));
    }
    _INST(address) = nil;
%}
! !

!ExternalBytes methodsFor:'printing & storing'!

displayString
    ^ 'ExternalBytes at:' , (address printStringRadix:16)
! !

!ExternalBytes methodsFor:'private accessing'!

setAddress:aNumber
    "set the start address"

    address := aNumber
! 

setAddress:aNumber size:sz 
    "set the start address and size"

    address := aNumber.
    size := sz
! 

setAddress:aNumber base:b size:sz 
    "set the start address, base address and size"

    address := aNumber.
    base := b.
    size := sz
! !

!ExternalBytes methodsFor:'accessing'!

address
    "return the start address"

    ^ address
!

base 
    "return the base address"

    base isNil ifTrue:[
        ^ address
    ].
    ^ base
!

basicSize
    "we do not know how many bytes are valid"

    size isNil ifTrue:[^ 0].
    ^ size
!

basicAt:index
    "return the byte at index, anInteger;
     this is the byte at (address + index - 1)"

%{  /* NOCONTEXT */

    unsigned char *cp;
    OBJ addr = _INST(address);

    if (__bothSmallInteger(addr, index)) {
        cp = (unsigned char *)(__intVal(addr));
        cp = cp + __intVal(index) - 1;
        RETURN ( _MKSMALLINT(*cp) );
    }
%}
.
    self primitiveFailed
!

basicAt:index put:value
    "set the byte at index, anInteger to value which must be 0..255"

%{  /* NOCONTEXT */

    unsigned char *cp;
    int val;
    OBJ addr = _INST(address);

    if (__isSmallInteger(addr)
     && __bothSmallInteger(index, value)) {
        val = __intVal(value);
        if ((val >= 0) && (val <= 255)) {
            cp = (unsigned char *)(__intVal(addr));
            cp = cp + __intVal(index) - 1;
            *cp = val;
            RETURN ( value );
        }
    }
%}
.
    self primitiveFailed
!

byteAt:index
    "return the unsigned byte at index, anInteger"

    ^ self basicAt:index
!

byteAt:index put:aByteValuedInteger
    "set the byte at index"

    ^ self basicAt:index put:aByteValuedInteger
!

wordAt:index
    "return the unsigned short at index, anInteger. fetching is MSB-first"

    ^ ((self at:index) bitShift:8) + (self at:(index + 1))
!

wordAt:index put:aNumber
    "set the short at index, index to aNumber; storing is MSB-first"

    self at:index       put:(aNumber digitAt:2).
    self at:(index + 1) put:(aNumber digitAt:1).
    ^ aNumber
!

doubleWordAt:index
    "return the unsigned long at index, anInteger. fetching is MSB-first"

    |val|

    val := self at:index.
    val := (val bitShift:8) + (self at:(index + 1)).
    val := (val bitShift:8) + (self at:(index + 2)).
    val := (val * 256) + (self at:(index + 3)).
    ^ val
!

doubleWordAt:index put:aNumber
    "set the long at index, index to aNumber; storing is MSB-first"


    self at:index       put:(aNumber digitAt:4).
    self at:(index + 1) put:(aNumber digitAt:3).
    self at:(index + 2) put:(aNumber digitAt:2).
    self at:(index + 3) put:(aNumber digitAt:1).
    ^ aNumber
! !
